#!/usr/bin/python3

import os
import pandas as pd
import requests
import scipy.sparse as sp
from .evaluation.d_utils import runner_utils as ru, eval_utils as eu
from .evaluation import config as c
from .evaluation.mappers import mapping_transformer as mt, gene_getter as gm, mapping_utils as mu
from .evaluation.mappers import disease_getter as dm
from .evaluation.mappers.mapper import Mapper, FileMapper
# only in full biodigest
import graph_tool as gt
import graph_tool.topology as gtt


def load_files(mapper: Mapper):
    """
    Run setup to load all needed files from api.

    :param mapper: object of type Mapper defining where and how to save the generated data
    """
    ru.print_current_usage('Starting Setup ...')

    ru.print_current_usage('Get id and attribute mappings ...')
    for file_id in mapper.loaded_mappings:
        r = requests.get(c.DIGEST+"name="+mapper.file_names[file_id])
        with open(os.path.join(mapper.files_dir, mapper.file_names[file_id]), 'wb') as f:
            f.write(r.content)

    ru.print_current_usage('Get distance mappings ...')
    for distance_measure in ["jaccard", "overlap"]:
        ru.print_current_usage('Get distance mappings for '+distance_measure+' ...')
        os.system("mkdir -p " + os.path.join(mapper.files_dir, distance_measure, ""))
        for file_id in mapper.loaded_distance_ids[distance_measure]:
            r = requests.get(c.DIGEST + "name=" + mapper.file_names[file_id] + "&measure="+distance_measure)
            with open(os.path.join(mapper.files_dir, distance_measure, mapper.file_names[file_id]), 'wb') as f:
                f.write(r.content)

        for file_id in mapper.loaded_distances[distance_measure]:
            r = requests.get(c.DIGEST + "name=" + mapper.file_names[file_id] + "&measure=" + distance_measure)
            with open(os.path.join(mapper.files_dir, distance_measure, mapper.file_names[file_id]), 'wb') as f:
                f.write(r.content)

    ru.print_current_usage('Get default networks ...')
    for network in ["ggi_graph.graphml", "ddi_graph.graphml"]:
        r = requests.get(c.DIGEST + "name=" + network)
        with open(os.path.join(mapper.files_dir, network), 'wb') as f:
            f.write(r.content)

    ru.print_current_usage('Finished Setup ...')


def create_files(mapper: Mapper):
    """
    Run setup to generate all files from scratch and gather data from databases.

    :param mapper: object of type Mapper defining where and how to save the generated data
    """
    ru.print_current_usage('Starting Setup ...')

    # ===== Start graph builder =====
    ggi = requests.post(url=c.NEDREX_GRAPH_BUILDER, json=c.NEDREX_GGI_POST)
    ddi = requests.post(url=c.NEDREX_GRAPH_BUILDER, json=c.NEDREX_DDI_POST)

    # ===== Load disorder ids =====
    ru.print_current_usage('Load NeDrEx disorder ids ...')
    disorder_ids = pd.read_csv(c.NEDREX_DISORDER_IDS, sep="\t")
    icd10_ids = pd.read_csv(c.NEDREX_ICD10_IDS, sep="\t")

    # ===== Transform disease mapping =====
    ru.print_current_usage('Transform loaded mappings ...')
    disease_mapping = mt.transform_id_mapping(disorder_ids.merge(icd10_ids, on='primaryDomainId', how='outer'))
    disease_mapping["ICD-10"] = disease_mapping["ICD-10"].apply(mt.reduce_to_parent)

    # ===== Save disease mapping =====
    mapper.update_mappings(in_df=disease_mapping, key='disorder_ids')

    # ===== Load gene ids =====
    ru.print_current_usage('Load NeDrEx gene ids ...')
    gene_ids = pd.read_csv(c.NEDREX_GENE_IDS, sep="\t")

    # ===== Transform gene mapping =====
    ru.print_current_usage('Transform loaded mappings ...')
    gene_ids = gene_ids.rename(columns={'primaryDomainId': "entrez"})
    gene_ids.entrez = gene_ids.entrez.str.replace(r'entrez.', '', regex=True)

    # ===== Get attributes =====
    ru.print_current_usage('Get gene attributes ...')
    gene_att_mapping = gm.get_gene_to_attributes(gene_set=gene_ids.entrez, id_type="entrez", mapper=mapper)

    ru.print_current_usage('Get disease attributes ...')
    # ===== Get DisGeNet attributes =====
    disgenet_mapping = pd.read_csv(c.DISGENET_DIS_MAP, compression='gzip', sep='\t', dtype=str)
    disgenet_mapping = disgenet_mapping[disgenet_mapping['vocabulary'] == 'MONDO'][['diseaseId', 'code']]
    disgenet_mapping = disgenet_mapping.rename(columns={'code': 'mondo'})

    var_mapping = mu.transform_disgenet_mapping(mapping=disgenet_mapping, file=c.DISGENET_REL_VARS, col_old='snpId',
                                                col_new='disgenet.variants_related_to_disease')
    gene_mapping = mu.transform_disgenet_mapping(mapping=disgenet_mapping, file=c.DISGENET_REL_GENES,
                                                 col_old='geneId',
                                                 col_new='disgenet.genes_related_to_disease')
    disease_att_mapping = pd.merge(var_mapping[['mondo', 'disgenet.variants_related_to_disease']],
                                   gene_mapping[['mondo', 'disgenet.genes_related_to_disease']],
                                   on="mondo", how="outer")

    # ===== Get KEGG attributes =====
    omim_to_hsa = pd.read_csv(c.KEGG_OMIM_TO_HSA, names=['hsa', 'omim', 'dir'], sep="\t", dtype=str)
    hsa_to_pathway = pd.read_csv(c.KEGG_HSA_TO_PATH, names=['hsa', 'pathway'], sep="\t", dtype=str)
    omim_to_pathway = pd.merge(omim_to_hsa[['hsa', 'omim']], hsa_to_pathway[['hsa', 'pathway']], on="hsa",
                               how="inner")
    omim_to_pathway.omim = omim_to_pathway.omim.str.replace('omim:', '')
    omim_to_pathway.pathway = omim_to_pathway.pathway.str.replace('path:', '')
    omim_to_pathway = pd.merge(disease_mapping[['mondo', 'omim']], omim_to_pathway[['omim', 'pathway']], on="omim",
                               how="inner")
    omim_to_pathway = omim_to_pathway[['mondo', 'pathway']].fillna('').groupby(['mondo'], as_index=False).agg(
        mu.combine_rows_to_set)
    omim_to_pathway.rename(columns={'pathway': 'ctd.pathway_related_to_disease'}, inplace=True)
    mapping = dm.get_attributes_from_database(
        missing=['MONDO:' + x for x in set(disease_att_mapping.mondo) - set(omim_to_pathway.mondo)],
        attributes=['ctd.pathway_related_to_disease'])
    mapping = pd.concat([omim_to_pathway, mapping])
    disease_att_mapping = pd.merge(disease_att_mapping, mapping[['mondo', 'ctd.pathway_related_to_disease']],
                                   on="mondo", how="left")

    # ===== Save mapping =====
    mapper.update_mappings(in_df=disease_att_mapping, key='disorder_atts')
    disease_att_mapping = dm.get_disease_to_attributes(disease_set=disease_mapping.mondo, id_type="mondo",
                                                       mapper=mapper)

    mapper.save_mappings()
    mapper.drop_mappings()

    # ===== Calculate pairwise comparisons =====
    ru.print_current_usage('Precalculate pairwise distances ...')

    for distance_measure in ["jaccard", "overlap"]:
        os.system("mkdir -p " + os.path.join(mapper.files_dir, distance_measure, ""))
        ru.print_current_usage('Precalculate pairwise distances for genes [' + distance_measure + '] ...')
        mapper.update_distance_ids(in_series=gene_att_mapping[c.ID_TYPE_KEY['entrez']], key='gene_mat_ids',
                                   distance_measure=distance_measure)
        for attribute in gene_att_mapping.columns[1:]:
            ru.print_current_usage('Precalculate pairwise distances for ' + attribute)
            subset_df = gene_att_mapping[gene_att_mapping[attribute].str.len() > 0]
            comp_mat = eu.get_distance_matrix(full_att_series=gene_att_mapping[attribute],
                                              from_ids=subset_df[c.ID_TYPE_KEY['entrez']],
                                              coefficient=distance_measure,
                                              id_to_index=mapper.loaded_distance_ids[distance_measure][
                                                  'gene_mat_ids'])
            sp.save_npz(os.path.join(mapper.files_dir, distance_measure, mapper.file_names[c.DISTANCES[attribute]]),
                        comp_mat)

        ru.print_current_usage('Precalculate pairwise distances for diseases [' + distance_measure + '] ...')
        mapper.update_distance_ids(in_series=disease_att_mapping['mondo'], key='disease_mat_ids',
                                   distance_measure=distance_measure)
        for attribute in disease_att_mapping.columns[1:]:
            ru.print_current_usage('Precalculate pairwise distances for ' + attribute)
            subset_df = disease_att_mapping[disease_att_mapping[attribute].str.len() > 0]
            comp_mat = eu.get_distance_matrix(full_att_series=disease_att_mapping[attribute],
                                              from_ids=subset_df['mondo'],
                                              coefficient=distance_measure,
                                              id_to_index=mapper.loaded_distance_ids[distance_measure][
                                                  'disease_mat_ids'])
            sp.save_npz(os.path.join(mapper.files_dir, distance_measure, mapper.file_names[c.DISTANCES[attribute]]),
                        comp_mat)

    mapper.save_distances()

    # ===== Calculate pairwise comparisons =====
    ru.print_current_usage('Create default networks ...')

    def recreate_network(graph_id, id_type, file_name, edges_num):
        response = requests.get(c.NEDREX_GRAPH_DOWNLOADER + graph_id.json() + ".graphml")
        open(os.path.join(mapper.files_dir, graph_id.json() + ".graphml"), "wb").write(response.content)
        graph = gt.load_graph(os.path.join(mapper.files_dir, graph_id.json() + ".graphml"))

        vertices = list()
        for v in graph.iter_vertices():
            if graph.vp['primaryDomainId'][v].startswith(id_type):
                vertices.append(graph.vertex(v))

        edges = list()
        for v in vertices:
            res = gtt.shortest_distance(graph, source=graph.vertex(v), max_dist=edges_num, target=vertices, directed=False)
            for index in range(len(res)):
                if res[index] == edges_num:
                    edges.append([graph.vp['primaryDomainId'][v],
                                  graph.vp['primaryDomainId'][vertices[index]]])

        edges = pd.DataFrame(edges, columns=["from", "to"])
        edges["from"] = edges["from"].str[len(id_type)+1:]
        edges["to"] = edges["to"].str[len(id_type)+1:]

        G = gt.Graph(directed=False)
        v_ids = G.add_edge_list(edges[["from", "to"]].values, hashed=True)
        G.vertex_properties['id'] = v_ids
        gt.stats.remove_parallel_edges(G)
        v_ids_set = set(v_ids)
        for v in vertices:
            cur_v = graph.vp['primaryDomainId'][v][len(id_type)+1:]
            if cur_v not in v_ids_set:
                new_v = G.add_vertex()
                G.vertex_properties['id'][new_v] = cur_v

        G.save(os.path.join(mapper.files_dir, file_name), fmt="graphml")
        os.system("rm " + os.path.join(mapper.files_dir, graph_id.json() + ".graphml"))

    recreate_network(graph_id=ggi, id_type="entrez", file_name="ggi_graph.graphml", edges_num=3)
    recreate_network(graph_id=ddi, id_type="mondo", file_name="ddi_graph.graphml", edges_num=2)

    ru.print_current_usage('Finished Setup ...')


def main(setup_type: str, replace: bool=True, path:str=c.FILES_DIR):
    os.system("mkdir -p " + os.path.join(path, "tmp", ""))
    if setup_type == "create":
        create_files(mapper=FileMapper(files_dir=os.path.join(path, "tmp", "")))
    elif setup_type == "api":
        load_files(mapper=FileMapper(files_dir=os.path.join(path, "tmp", "")))
    if replace:
        os.system("cp -r " + os.path.join(path, "tmp", "") + "* " + path)
        os.system("rm -rf " + os.path.join(path, "tmp", ""))


if __name__ == "__main__":
    desc = "     Run setup to create/load precalculated files."
    args = ru.save_parameters(script_desc=desc, arguments=('s'))
    main(setup_type=args.setup_type)



#!/usr/bin/python3

import pandas as pd
import numpy as np
from pathlib import Path
from abc import abstractmethod
from .. import config
from . import mapping_utils as mu
import scipy.sparse as sp
import pickle
import os


class Mapper:
    loaded_mappings = {'gene_ids': pd.DataFrame(), 'gene_atts': pd.DataFrame(), 'disorder_ids': pd.DataFrame(),
                       'disorder_atts': pd.DataFrame()}

    loaded_distance_ids = {'jaccard': {'gene_mat_ids': dict(), 'disease_mat_ids': dict()},
                           'overlap': {'gene_mat_ids': dict(), 'disease_mat_ids': dict()}}
    loaded_distances = {
        'jaccard': {'go_BP': sp.csr_matrix(None), 'go_CC': sp.csr_matrix(None), 'go_MF': sp.csr_matrix(None),
                    'pathway_kegg': sp.csr_matrix(None), 'related_genes': sp.csr_matrix(None),
                    'related_variants': sp.csr_matrix(None), 'related_pathways': sp.csr_matrix(None)},
        'overlap': {'go_BP': sp.csr_matrix(None), 'go_CC': sp.csr_matrix(None), 'go_MF': sp.csr_matrix(None),
                    'pathway_kegg': sp.csr_matrix(None), 'related_genes': sp.csr_matrix(None),
                    'related_variants': sp.csr_matrix(None), 'related_pathways': sp.csr_matrix(None)}}

    changed_mappings = set()

    load: bool = True

    def __init__(self, preload: bool = False):
        if preload:
            self.load_mappings()
            self.load_distances(set_type="entrez", distance_measure="jaccard")
            self.load_distances(set_type="mondo", distance_measure="jaccard")
            self.load_distances(set_type="entrez", distance_measure="overlap")
            self.load_distances(set_type="mondo", distance_measure="overlap")
            self.load = False

    @abstractmethod
    def load_mappings(self):
        pass

    @abstractmethod
    def load_distances(self, set_type: str, distance_measure: str):
        pass

    @abstractmethod
    def load_file(self, key: str, in_type: str):
        pass

    def get_loaded_mapping(self, in_set, id_type: str, key: str):
        if not self.loaded_mappings[key].empty:
            current_mapping = self.loaded_mappings[key].copy()
            if id_type not in ['entrezgene', 'mondo']:
                current_mapping = current_mapping.explode(id_type)
            hit_mapping = current_mapping.loc[current_mapping[id_type].isin(in_set)]
            if not hit_mapping.empty:
                return hit_mapping, set(in_set) - set(hit_mapping[id_type])
        return pd.DataFrame(), in_set

    def update_mappings(self, in_df: pd.DataFrame(), key: str):
        self.changed_mappings.add(key)
        if not self.loaded_mappings[key].empty:
            self.loaded_mappings[key] = pd.concat([self.loaded_mappings[key], in_df], ignore_index=True)
        else:
            self.loaded_mappings[key] = in_df

    def get_loaded_mapping_ids(self, in_ids, id_type: str) -> pd.DataFrame:
        if id_type in config.SUPPORTED_GENE_IDS:
            mapping, _ = self.get_loaded_mapping(in_set=in_ids, id_type=config.ID_TYPE_KEY[id_type], key="gene_ids")
            return mapping
        else:  # if set_type in config.SUPPORTED_DISEASE_IDS
            mapping, _ = self.get_loaded_mapping(in_set=in_ids, id_type=config.ID_TYPE_KEY[id_type], key="disorder_ids")
            return mapping

    def update_distance_ids(self, in_series: pd.Series, key: str, distance_measure: str) -> pd.Series:
        self.changed_mappings.add(distance_measure + "_" + key)
        if self.loaded_distance_ids[distance_measure][key]:  # is not empty
            if len(self.loaded_distance_ids[distance_measure][key].keys()) < len(in_series):
                new_ids = in_series[len(self.loaded_distance_ids[distance_measure][key].keys()):]
                for index, value in enumerate(iterable=new_ids,
                                              start=len(self.loaded_distance_ids[distance_measure][key].keys())):
                    self.loaded_distance_ids[distance_measure][key][value] = index
                return new_ids
            else:
                return pd.Series([],dtype=pd.StringDtype())
        else:
            for index, value in enumerate(iterable=in_series, start=0):
                self.loaded_distance_ids[distance_measure][key][value] = index
            return in_series

    def get_loaded_distances(self, in_series: pd.Series, id_type: str, key: str, distance_measure: str,
                             to_series: pd.Series = None) -> sp.csr_matrix:
        if self.loaded_distance_ids[distance_measure][id_type]:  # is not empty
            indices = list()
            for element in in_series:
                indices.append(self.loaded_distance_ids[distance_measure][id_type][element])
            if to_series is not None:
                to_indices = list()
                for element in to_series:
                    to_indices.append(self.loaded_distance_ids[distance_measure][id_type][element])
                return self.loaded_distances[distance_measure][key][indices, :][:, to_indices]
            else:
                return self.loaded_distances[distance_measure][key][indices, :][:, indices]
        else:
            return sp.csr_matrix(None)

    def update_distances(self, in_mat: sp.coo_matrix, id_type: str, key: str, distance_measure: str):
        self.changed_mappings.add(distance_measure + "_" + key)
        if self.loaded_distances[distance_measure][key].nnz > 0:
            old_mat = self.loaded_distances[distance_measure][key].tocoo()
            row = np.concatenate((old_mat.row, in_mat.row), axis=None)
            col = np.concatenate((old_mat.col, in_mat.col), axis=None)
            data = np.concatenate((old_mat.data, in_mat.data), axis=None)
            self.loaded_distances[distance_measure][key] = sp.csr_matrix((data, (row, col)), shape=(
                len(self.loaded_distance_ids[distance_measure][id_type].keys()),
                len(self.loaded_distance_ids[distance_measure][id_type].keys())))
        else:
            self.loaded_distances[distance_measure][key] = in_mat.tocsr()

    def get_full_set(self, id_type: str, mapping_name: str) -> pd.DataFrame:
        current_mapping = self.loaded_mappings[mapping_name]
        if id_type not in ['entrezgene', 'mondo']:
            current_mapping = current_mapping.explode(id_type).fillna("")
        return current_mapping

    @abstractmethod
    def save_mappings(self):
        pass

    @abstractmethod
    def save_distances(self):
        pass

    @abstractmethod
    def save_file(self, in_object, key: str, in_type: str):
        pass

    @abstractmethod
    def check_for_setup_sources(self):
        pass

    def drop_mappings(self):
        self.loaded_mappings = {'gene_ids': pd.DataFrame(), 'gene_atts': pd.DataFrame(), 'disorder_ids': pd.DataFrame(),
                                'disorder_atts': pd.DataFrame()}


class FileMapper(Mapper):
    file_names = {'gene_ids': 'gene_id_mapping.csv',
                  'gene_atts': 'gene_att_mapping.csv',
                  'disorder_ids': 'disease_id_mapping.csv',
                  'disorder_atts': 'disease_att_mapping.csv',
                  'gene_mat_ids': 'gene_mat_ids.pkl',
                  'disease_mat_ids': 'disease_mat_ids.pkl',
                  'go_BP': 'gene_dist_go_BP.npz',
                  'go_CC': 'gene_dist_go_CC.npz',
                  'go_MF': 'gene_dist_go_MF.npz',
                  'pathway_kegg': 'gene_dist_pathway_kegg.npz',
                  'related_genes': 'disease_dist_rel_genes.npz',
                  'related_variants': 'disease_dist_rel_variants.npz',
                  'related_pathways': 'disease_dist_rel_pathways.npz'}

    def __init__(self, preload: bool = False, files_dir=config.FILES_DIR):
        self.files_dir = files_dir
        super().__init__(preload=preload)

    def load_mappings(self):
        if self.load:
            for mapping_key in ['gene_atts', 'disorder_atts', 'gene_ids', 'disorder_ids']:
                self.load_file(key=mapping_key, in_type='mapping')
                self.loaded_mappings[mapping_key][self.loaded_mappings[mapping_key].columns[1:]] = \
                    self.loaded_mappings[mapping_key][self.loaded_mappings[mapping_key].columns[1:]].fillna(
                        '').applymap(mu.string_to_set)

    def _load_file_mapping(self, file, sep, mapping_name):
        """
        Get previous mappings from file.

        :param file: file with previous distances
        :param sep: separator of values in file
        :param mapping_name: key name of local dictionary for saving data
        :return: dataframe with previous mapping
        """
        # ===== Get mapping from local mapping file =====
        mapping = pd.read_csv(file, sep=sep, header=0, dtype=str).fillna('')
        if mapping_name == "disorder_ids":
            icd_unstack = mu.split_and_expand_column(data=mapping, split_string=",", column_name="ICD-10")
            mapping = pd.concat([icd_unstack, mapping[mapping['ICD-10'] != '']])
        # ===== Save mapping to local dictionary =====
        self.loaded_mappings[mapping_name] = mapping

    def load_distances(self, set_type: str, distance_measure: str):
        if self.load:
            if set_type in config.SUPPORTED_GENE_IDS:
                self.load_file(key='gene_mat_ids', in_type='gene_id', distance_measure=distance_measure)
                for distance_key in ['go_BP', 'go_CC', 'go_MF', 'pathway_kegg']:
                    self.load_file(key=distance_key, in_type='distance', distance_measure=distance_measure)
            else:  # if set_type in config.SUPPORTED_DISEASE_IDS
                self.load_file(key='disease_mat_ids', in_type='distance_id', distance_measure=distance_measure)
                for distance_key in ['related_genes', 'related_variants', 'related_pathways']:
                    self.load_file(key=distance_key, in_type='distance', distance_measure=distance_measure)

    def load_file(self, key: str, in_type: str, distance_measure: str = None):
        if in_type == "mapping":
            self._load_file_mapping(file=os.path.join(self.files_dir, self.file_names[key]), sep=",", mapping_name=key)
        elif in_type == "distance":
            self.loaded_distances[distance_measure][key] = sp.load_npz(
                os.path.join(self.files_dir, distance_measure, self.file_names[key])).tocsr()
        else:  # in_type == "distance_id"
            with open(os.path.join(self.files_dir, distance_measure, self.file_names[key]), 'rb') as f:
                self.loaded_distance_ids[distance_measure][key] = pickle.load(f)

    def save_mappings(self):
        for mapping_key in ['gene_ids', 'disorder_ids']:
            if not self.loaded_mappings[mapping_key].empty and mapping_key in self.changed_mappings:
                self.save_file(in_object=self.loaded_mappings[mapping_key], key=mapping_key, in_type='mapping')
        for mapping_key in ['gene_atts', 'disorder_atts']:
            if not self.loaded_mappings[mapping_key].empty and mapping_key in self.changed_mappings:
                df = self.loaded_mappings[mapping_key]
                df[df.columns[1:]] = df[df.columns[1:]].fillna('').applymap(mu.set_to_string)
                self.save_file(in_object=df, key=mapping_key, in_type='mapping')

    def save_distances(self):
        for distance_measure in ['jaccard', 'overlap']:
            os.system("mkdir -p " + os.path.join(self.files_dir, distance_measure))
            for distance_id_key in ['gene_mat_ids', 'disease_mat_ids']:
                if distance_measure + "_" + distance_id_key in self.changed_mappings:
                    self.save_file(in_object=self.loaded_distance_ids[distance_measure][distance_id_key],
                                   key=distance_id_key, distance_measure=distance_measure,
                                   in_type='distance_id')
            for distance_key in ['go_BP', 'go_CC', 'go_MF', 'pathway_kegg', 'related_genes', 'related_variants',
                                 'related_pathways']:
                if distance_measure + "_" + distance_key in self.changed_mappings:
                    self.save_file(in_object=self.loaded_distances[distance_measure][distance_key], key=distance_key,
                                   in_type='distance', distance_measure=distance_measure)

    def save_file(self, in_object, key: str, in_type: str, distance_measure: str = None):
        if in_type == "mapping":
            if not self.loaded_mappings[key].empty:
                in_object.to_csv(os.path.join(self.files_dir, self.file_names[key]), index=False)
        elif in_type == "distance":
            if self.loaded_distances[distance_measure][key].nnz > 0:
                sp.save_npz(os.path.join(self.files_dir, distance_measure, self.file_names[key]), in_object.tocoo())
        else:  # in_type == "distance_id"
            if self.loaded_distance_ids[distance_measure][key]:
                with open(os.path.join(self.files_dir, distance_measure, self.file_names[key]), 'wb+') as f:
                    pickle.dump(in_object, f)

    def check_for_setup_sources(self):
        for key in ['gene_atts', 'disorder_atts', 'gene_ids', 'disorder_ids']:
            if not Path(os.path.join(self.files_dir, self.file_names[key])).is_file():
                raise Exception(self.file_names[key] + "does not exist. Please run setup.")
        for distance_measure in ['jaccard', 'overlap']:
            for key in ['gene_mat_ids', 'disease_mat_ids', 'go_BP', 'go_CC', 'go_MF', 'pathway_kegg',
                        'related_genes', 'related_variants', 'related_pathways']:
                if not Path(os.path.join(self.files_dir, distance_measure, self.file_names[key])).is_file():
                    raise Exception(self.file_names[key] + "does not exist. Please run setup.")



#!/usr/bin/python3

import os
import numpy as np
import pandas as pd
from .evaluation.d_utils import runner_utils as ru, eval_utils as eu, plotting_utils as pu
from .evaluation.mappers.mapper import Mapper, FileMapper
from .evaluation import config, comparator as comp, background_models as bm
import random
import json
import time
from pathlib import Path
from typing import Union, Callable


def single_validation(tar: Union[pd.DataFrame, set], tar_id: str, mode: str, distance: str = "jaccard",
                      ref: set = None, ref_id: str = None, enriched: bool = False,
                      mapper: Mapper = FileMapper(), runs: int = config.NUMBER_OF_RANDOM_RUNS,
                      background_model: str = "complete", replace=100, verbose: bool = False,
                      network_data: dict = None,
                      progress: Callable[[float, str], None] = None):
    """
    Single validation of a set, cluster a id versus set and set versus set.

    :param tar: cluster as type dataframe with columns=["id","cluster"] or set as type set
    :param tar_id: id type of target input
    :param mode: comparison mode [set, set-set, clustering, subnetwork, subnetwork-set]
    :param distance: distance measure used for comparison. [Default=jaccard]
    :param ref: set of reference ids [Default=None]
    :param ref_id: id type of reference input [Default=None]
    :param enriched: bool setting if values of reference set should be filtered for enriched values [Default=False]
    :param mapper: mapper from type Mapper defining where the precalculated information comes from [Default=FileMapper]
    :param runs: number of random runs to create p-values [Default=1000]
    :param background_model: which background model to use for random picks [Default="complete"]
    :param replace: how many % of target input should be replaced by random picks [Default=100]
    :param verbose: bool if additional info like ids without assigned attributes should be printed [Default=False]
    :param network_data: dict consisting of {"network_file": path to network file,
    "prop_name": name of vertex property with ids if network file of type graphml or gt,
    "id_type": id type of network ids}
    :param progress: method that will get a float [0,1] and message indicating the current progress
    """
    ru.start_time = time.time()
    ru.print_current_usage('Check for proper setup ...') if verbose else None
    mapper.check_for_setup_sources()
    # ===== Comparison with a set =====
    ru.print_current_usage('Starting validation ...')
    progress(0.01, "Load mappings...") if progress is not None else None
    if mapper.load:
        ru.print_current_usage('Load mappings for input into cache ...') if verbose else None
        mapper.load_mappings()
    if mode in ["set", "set-set", "subnetwork", "subnetwork-set"]:
        error_mappings = []
        if mode in ["set-set", "subnetwork-set"]:
            comparator = comp.SetSetComparator(mapper=mapper, enriched=enriched, verbose=verbose,
                                               distance_measure=distance)
            comparator.load_reference(ref=ref, ref_id_type=ref_id, tar_id_type=tar_id)
            if not comparator.ref_dict:
                error_mappings.append("reference")
        else:  # mode == "set"
            comparator = comp.SetComparator(mapper=mapper, verbose=verbose, distance_measure=distance)
            if mapper.load:
                ru.print_current_usage('Load distances for input into cache ...') if verbose else None
                progress(0.05, "Load distances...") if progress is not None else None
                mapper.load_distances(set_type=tar_id, distance_measure=distance)
        comparator.load_target(id_set=tar, id_type=tar_id)
        # ===== Check if mappings possible =====
        if comparator.mapping.empty:
            error_mappings.append("target set")
        if len(error_mappings) > 0:
            ru.print_current_usage(
                'Exit validation: No mapping found for ' + ' and '.join(error_mappings)) if verbose else None
            return {'status': 'No mapping found for ' + ' and '.join(error_mappings),
                    'input_values': {'values': None, 'mapped_ids': []}, 'random_values': None,
                    'p_values': {'values': None}}
        # ===== Get validation values of input =====
        ru.print_current_usage('Validation of input ...') if verbose else None
        progress(0.1, "Validation of input...") if progress is not None else None
        my_value, mapped = comparator.compare()
        ru.print_current_usage('Validation of random runs ...') if verbose else None
        progress(0.1 + (0.9 / min(runs + 1, 100)),
                 "Validation with background model...") if progress is not None else None
        comparator.verbose = False
        comparator.input_run = False
        comp_values = get_random_runs_values(comparator=comparator, mode=mode, mapper=mapper, tar_id=tar_id,
                                             runs=runs, background_model=background_model, replace=replace,
                                             network_data=network_data, progress=progress)
        # ===== Statistical analysis =====
        ru.print_current_usage('Calculating p-values ...') if verbose else None
        set_value = eu.calc_pvalue(test_value=my_value, random_values=comp_values[0], maximize=True)
        measure_short = {"jaccard": "JI-based", "overlap": "OC-based"}
        p_values = {measure_short[distance]: set_value}
        results = {'status': 'ok',
                   'input_values': {'values': {measure_short[distance]: my_value}, 'mapped_ids': mapped},
                   'random_values': {'values': {measure_short[distance]: comp_values[0].to_dict('list')}},
                   'p_values': {'values': p_values}}

    # ===== Special case cluster =====
    elif mode == "clustering":
        if mapper.load:
            ru.print_current_usage('Load distances for input into cache ...') if verbose else None
            progress(0.05, "Load distances...") if progress is not None else None
            mapper.load_distances(set_type=tar_id, distance_measure=distance)
        ru.print_current_usage('Load input data ...') if verbose else None
        comparator = comp.ClusterComparator(mapper=mapper, verbose=verbose, distance_measure=distance)
        comparator.load_target(id_set=tar, id_type=tar_id)
        # ===== Check if mappings possible =====
        if comparator.mapping.empty:
            ru.print_current_usage(
                'Exit validation: No mapping found for target cluster') if verbose else None
            return {'status': 'No mapping found for target cluster',
                    'input_values': {'values': None, 'mapped_ids': []},
                    'random_values': {'values': None}, 'p_values': {'values': None}}
        # ===== Get validation values of input =====
        ru.print_current_usage('Validation of input ...') if verbose else None
        progress(0.1, "Validation of input...") if progress is not None else None
        my_value_di, my_value_ss, my_value_dbi, my_value_ss_inter, mapped = comparator.compare()
        ru.print_current_usage('Validation of random runs ...') if verbose else None
        progress(0.1 + (0.9 / min(runs + 1, 100)),
                 "Validation with background model...") if progress is not None else None
        comparator.verbose = False
        comparator.input_run = False
        comp_values = get_random_runs_values(comparator=comparator, mode=mode, mapper=mapper, tar_id=tar_id,
                                             runs=runs, progress=progress)
        # ===== Statistical analysis =====
        ru.print_current_usage('Calculating p-values ...') if verbose else None
        p_values_di = eu.calc_pvalue(test_value=my_value_di, random_values=comp_values[0], maximize=False)
        p_values_ss = eu.calc_pvalue(test_value=my_value_ss, random_values=comp_values[1], maximize=True)
        p_values_dbi = eu.calc_pvalue(test_value=my_value_dbi, random_values=comp_values[2],
                                      maximize=False)
        p_values = {'DI-based': p_values_di, 'SS-based': p_values_ss, 'DBI-based': p_values_dbi}
        results = {'status': 'ok',
                   'input_values': {'values': {'DI-based': my_value_di, 'SS-based': my_value_ss,
                                               'DBI-based': my_value_dbi},
                                    'values_inter': my_value_ss_inter,
                                    'mapped_ids': mapped},
                   'random_values': {'values': {'DI-based': comp_values[0].to_dict('list'),
                                                'SS-based': comp_values[1].to_dict('list'),
                                                'DBI-based': comp_values[2].to_dict('list')}},
                   'p_values': {'values': p_values}}
    else:
        results = {None}
    ru.print_current_usage('Finished validation')
    return results


def get_random_runs_values(comparator: comp.Comparator, mode: str, mapper: Mapper, tar_id: str, runs: int,
                           background_model: str = "complete", replace=100, term: str = "sum",
                           network_data: dict = None, progress: Callable[[float, str], None] = None) -> list:
    """
    Pick random ids to recreate a target input and run the comparison against reference or itself.
    The random ids are of the same id type of the original target input.

    :param comparator: comparator with defined way how to make the comparison
    :param mode: comparison mode [set, id-set, set-set, cluster]
    :param mapper: mapper from type Mapper defining where the precalculated information comes from
    :param tar_id: id type of target input
    :param runs: number of random runs to create p-values [Default=1000]
    :param background_model: which background model to use for random picks [Default="complete"]
    :param replace: how many % of target input should be replaced by random picks [Default=100]
    :param term: on what term the term preserving background model should calculate [Default="sum"]
    :param network_data: dict consisting of {"network_file": path to network file,
    "prop_name": name of vertex property with ids if network file of type graphml or gt,
    "id_type": id type of network ids}
    :param progress: method that will get a float [0,1] and message indicating the current progress
    :return: comparison
    """
    results = list()
    threshold = min(runs + 1, 100)
    limit, counter = int((runs + 1) / threshold), 1
    if not mode == "clustering":
        results.extend([list()])
        # ===== Get full id mapping =====
        if tar_id in config.SUPPORTED_DISEASE_IDS:
            full_id_map = mapper.get_full_set(id_type=tar_id, mapping_name='disorder_ids')
            new_id_type, map_id_type, map_att_type = "mondo", "disorder_ids", "disorder_atts"
        else:  # if tar_id in config.SUPPORTED_GENE_IDS:
            full_id_map = mapper.get_full_set(id_type=config.ID_TYPE_KEY[tar_id], mapping_name='gene_ids')
            new_id_type, map_id_type, map_att_type = "entrez", "gene_ids", "gene_atts"
        # ===== Precalculate sizes =====
        orig_ids = set(comparator.id_set)
        size = len(orig_ids)
        random_size = int((size / 100) * replace)

        if background_model == "complete":
            background = bm.CompleteModel(prev_id_type=tar_id, full_id_map=full_id_map)
        elif background_model == "term-pres":
            background = bm.TermPresModel(mapper=mapper, prev_id_type=tar_id, new_id_type=new_id_type,
                                          map_id_type=map_id_type, map_att_type=map_att_type, term=term)
        elif background_model == "network":
            if network_data is None:  # load default network from config
                if new_id_type == "mondo":
                    network_data = {"network_file": os.path.join(mapper.files_dir, "ddi_graph.graphml"),
                                    "id_type": new_id_type, "prop_name": "id"}
                else:
                    network_data = {"network_file": os.path.join(mapper.files_dir, "ggi_graph.graphml"),
                                    "id_type": new_id_type, "prop_name": "id"}
            if network_data["id_type"] != tar_id:  # remap ids
                input_ids = set(full_id_map[full_id_map[config.ID_TYPE_KEY[tar_id]].isin(orig_ids)][
                                    config.ID_TYPE_KEY[network_data["id_type"]]])
                background = bm.NetworkModel(network_data=network_data, to_replace=input_ids, N=runs)
            else:
                background = bm.NetworkModel(network_data=network_data, to_replace=orig_ids, N=runs)

        else:
            return list()
        for run in range(1, runs + 1):
            # ===== Update progress =====
            if run % limit == 0:
                progress(0.1 + (counter * (0.9 / threshold)),
                         str(
                             limit * counter) + " run(s) with background model finished...") if progress is not None else None
                counter += 1
            # ===== Pick new samples =====
            old_sample = set(random.sample(list(orig_ids), (size - random_size)))  # ignore when network model
            to_replace = orig_ids.difference(old_sample)

            if background_model == "complete":
                random_sample = background.get_module(to_replace=to_replace)
            elif background_model == "term-pres":
                random_sample = background.get_module(to_replace=to_replace, term=term, prev_id_type=tar_id)
            elif background_model == "network":
                random_sample = background.get_module(index=run - 1)
                if network_data["id_type"] != tar_id:  # remap ids
                    # random_sample = \
                    # set(full_id_map[full_id_map[config.ID_TYPE_KEY[network_data["id_type"]]].isin(random_sample)][
                    #    config.ID_TYPE_KEY[tar_id]])
                    tar_id = network_data["id_type"]
                    old_sample = set()
            else:
                return list()

            # ===== Get corresponding id set =====
            id_set = full_id_map[full_id_map[config.ID_TYPE_KEY[tar_id]].isin(random_sample.union(old_sample))][
                comparator.att_id]
            # ===== Calculate values =====
            comparator.load_target(id_set=set(id_set), id_type=new_id_type)
            result, _ = comparator.compare()
            if result:
                results[0].append(result)

    # ===== Special case cluster =====
    else:
        # ===== Calculate values =====
        results.extend([list(), list(), list()])
        # ===== Precalculate sizes =====
        orig_ids = set(comparator.clustering['id'])
        orig_clusters = comparator.clustering
        size = len(orig_ids)
        random_size = int((size / 100) * replace)
        for run in range(0, runs):
            # ===== Update progress =====
            if run % limit == 0:
                progress(0.1 + (counter * (0.9 / threshold)),
                         str(
                             limit * counter) + " run(s) with background model finished...") if progress is not None else None
                counter += 1
            old_sample = set(random.sample(orig_ids, (size - random_size)))
            # ===== Shuffle subset of clusters =====
            subset = orig_clusters[~orig_clusters['id'].isin(old_sample)]
            subset["cluster_index"] = np.random.permutation(subset["cluster_index"])
            comparator.clustering = pd.concat([orig_clusters[orig_clusters['id'].isin(old_sample)], subset])
            # ===== Start validating =====
            value_di, value_ss, value_dbi, value_ss_inter, mapped = comparator.compare()
            results[0].append(value_di)
            results[1].append(value_ss)
            results[2].append(value_dbi)
    # convert to dataframes
    final_results = list()
    for result in results:
        final_results.append(pd.DataFrame(result))
    return final_results


def save_results(results: dict, prefix: str, out_dir):
    # ===== Save complete output =====
    with open(os.path.join(out_dir, prefix + "_result.json"), "w") as outfile:
        json.dump(results, outfile)
    # ===== Save output to tables =====
    if results["status"] == "ok":
        pd.DataFrame(results["input_values"]['values']).to_csv(os.path.join(out_dir, prefix + "_input_validation.csv"))
        pd.DataFrame(results["p_values"]['values']).to_csv(os.path.join(out_dir, prefix + "_p-value_validation.csv"))


def significance_contribution(results: dict, excluded: str, tar: Union[pd.DataFrame, set], tar_id: str, mode: str,
                              distance: str = "jaccard", ref: set = None, ref_id: str = None, enriched: bool = False,
                              mapper: Mapper = FileMapper(), runs: int = config.NUMBER_OF_RANDOM_RUNS,
                              background_model: str = "complete",
                              replace=100, verbose: bool = False, network_data: dict = None):
    """
    Calculate the significance contribution of the excluded id by removing it from the target set,
    running the single_validation on the subset and then comparing the empirical p-values to determine, if
    including the id in the original set would create a positive or negative contribution.

    :param results: results from the single_validation run to compare to
    :param excluded: id of which the significance contribution should be calculated for
    :param tar: cluster as type dataframe with columns=["id","cluster"] or set as type set
    :param tar_id: id type of target input
    :param mode: comparison mode [set, set-set, clustering, subnetwork, subnetwork-set]
    :param distance: distance measure used for comparison. [Default=jaccard]
    :param ref: set of reference ids [Default=None]
    :param ref_id: id type of reference input [Default=None]
    :param enriched: bool setting if values of reference set should be filtered for enriched values [Default=False]
    :param mapper: mapper from type Mapper defining where the precalculated information comes from [Default=FileMapper]
    :param runs: number of random runs to create p-values [Default=1000]
    :param background_model: which background model to use for random picks [Default="complete"]
    :param replace: how many % of target input should be replaced by random picks [Default=100]
    :param verbose: bool if additional info like ids without assigned attributes should be printed [Default=False]
    :param network_data: dict consisting of {"network_file": path to network file,
    "prop_name": name of vertex property with ids if network file of type graphml or gt,
    "id_type": id type of network ids}
    """
    if results["status"] != "ok":
        return dict()
    # ===== Create new subset =====
    if isinstance(tar, set):
        new_tar = tar.copy()
        new_tar.remove(excluded)
    elif isinstance(tar, pd.Series):
        tar.index = tar
        new_tar = tar.drop(labels=[excluded]).reset_index(drop=True)
    elif isinstance(tar, list):
        new_tar = tar.copy()
        new_tar.remove(excluded)
    else:  # isinstance(tar, pd.DataFrame):
        new_tar = tar[tar[tar.columns[0]] != excluded]
    # ===== Run single validation on new subset =====
    results_sig = single_validation(tar=new_tar, tar_id=tar_id, ref=ref, ref_id=ref_id,
                                    mode=mode, mapper=mapper, runs=runs,
                                    background_model=background_model, verbose=verbose,
                                    enriched=enriched, replace=replace,
                                    distance=distance, network_data=network_data)
    # ===== Calculate significance contribution =====
    new_df = pd.DataFrame(results_sig["p_values"]['values']) - pd.DataFrame(results["p_values"]['values'])
    return {excluded: new_df.to_dict()}


def significance_contributions(results: dict, tar: Union[pd.DataFrame, set], tar_id: str, mode: str,
                               distance: str = "jaccard", ref: set = None, ref_id: str = None, enriched: bool = False,
                               mapper: Mapper = FileMapper(), runs: int = config.NUMBER_OF_RANDOM_RUNS,
                               background_model: str = "complete",
                               replace=100, verbose: bool = False, network_data: dict = None,
                               progress: Callable[[float, str], None] = None):
    """
    Calculate the significance contribution of every input id separately by removing it from the target set,
    running the single_validation on the subset and then comparing the empirical p-values to determine, if
    including the id in the original set would create a positive or negative contribution.

    :param results: results from the single_validation run to compare to
    :param tar: cluster as type dataframe with columns=["id","cluster"] or set as type set
    :param tar_id: id type of target input
    :param mode: comparison mode [set, set-set, clustering, subnetwork, subnetwork-set]
    :param distance: distance measure used for comparison. [Default=jaccard]
    :param ref: set of reference ids [Default=None]
    :param ref_id: id type of reference input [Default=None]
    :param enriched: bool setting if values of reference set should be filtered for enriched values [Default=False]
    :param mapper: mapper from type Mapper defining where the precalculated information comes from [Default=FileMapper]
    :param runs: number of random runs to create p-values [Default=1000]
    :param background_model: which background model to use for random picks [Default="complete"]
    :param replace: how many % of target input should be replaced by random picks [Default=100]
    :param verbose: bool if additional info like ids without assigned attributes should be printed [Default=False]
    :param network_data: dict consisting of {"network_file": path to network file,
    "prop_name": name of vertex property with ids if network file of type graphml or gt,
    "id_type": id type of network ids}
    :param progress: method that will get a float [0,1] and message indicating the current progress
    """
    if results["status"] != "ok":
        return dict()
    results_sig = dict()
    progress(0.05, "Prepare for run ...") if progress is not None else None
    # ===== Calculate significance for each id from input set =====
    for index, excluded in enumerate(tar):
        result_sig = significance_contribution(results=results, excluded=excluded, tar=tar, tar_id=tar_id, ref=ref,
                                               ref_id=ref_id, mode=mode, mapper=mapper, runs=runs,
                                               background_model=background_model, verbose=False, enriched=enriched,
                                               replace=replace, distance=distance, network_data=network_data)
        results_sig.update(result_sig)
        ru.print_current_usage(
            '{}/{} significance contributions calculated ...'.format(index + 1, len(tar))) if verbose else None
        progress(0.05 + ((0.95 / len(tar)) * (index + 1)), str(index + 1) + "/" + str(
            len(tar)) + " significance contributions calculated ...") if progress is not None else None
    # ===== Transform dict for later plot use =====
    # id -> stat_type -> values to stat_type -> id -> values
    final_results_sig = transform_dict(results_sig)
    return final_results_sig


def transform_dict(in_dict):
    """
    Transform dictionary from significance calculation.
    id -> stat_type -> values to stat_type -> id -> values

    :param in_dict: result dict from significance_calculations
    :return: transformed dictionary
    """
    if not in_dict:  # if empty
        return in_dict
    out_dict = dict()
    for a in in_dict:
        for b in in_dict[a]:
            if b not in out_dict:
                out_dict[b] = dict()
            out_dict[b][a] = in_dict[a][b]
    return out_dict


def save_contribution_results(results: dict, prefix: str, out_dir):
    # ===== Save complete output =====
    with open(os.path.join(out_dir, prefix + "_contribution_result.json"), "w") as outfile:
        json.dump(results, outfile)
    # ===== Save output to tables =====
    if results:  # not empty
        for stat_type in results:
            pd.DataFrame(results[stat_type]).T.to_csv(
                os.path.join(out_dir, prefix + "_" + stat_type + "_contribution.csv"))


if __name__ == "__main__":
    desc = "            Evaluation of disease and gene sets, clusterings or subnetworks."
    args = ru.save_parameters(script_desc=desc,
                              arguments=('r', 'ri', 't', 'ti', 'm', 'o', 'e', 'c', 'v', 'b', 'pr', 'p', 'dg',
                                         'n', 'ni', 'np', 'sc'))
    mapper = FileMapper()
    res = single_validation(tar=args.target, tar_id=args.target_id_type, verbose=args.verbose,
                            mode=args.mode, ref=args.reference, ref_id=args.reference_id_type, mapper=mapper,
                            enriched=args.enriched, runs=args.runs, distance=args.distance_measure,
                            background_model=args.background_model, replace=args.replace,
                            network_data={"network_file": args.network, "prop_name": args.network_property_name,
                                          "id_type": args.network_id_type})
    # ===== Saving final files and results =====
    ru.print_current_usage('Save files') if args.verbose else None
    Path(args.out_dir).mkdir(parents=True, exist_ok=True)  # make sure output dir exists
    pref = "_".join([args.mode, args.target_id_type, args.background_model])
    save_results(results=res, prefix=pref, out_dir=args.out_dir)
    if args.plot and res["status"] == "ok":
        pu.create_plots(results=res, mode=args.mode, tar=args.target, tar_id=args.target_id_type,
                        out_dir=args.out_dir, prefix=pref)
        pu.create_extended_plots(results=res, mode=args.mode, mapper=mapper, tar=args.target,
                                 out_dir=args.out_dir, prefix=pref)
    if res["status"] == "ok":
        ru.print_current_usage('Calculation of significance contribution') if args.verbose else None
        if args.significance_contribution:
            res_sig = significance_contributions(results=res, tar=args.target, tar_id=args.target_id_type,
                                                 verbose=args.verbose, mode=args.mode, ref=args.reference,
                                                 ref_id=args.reference_id_type, mapper=mapper,
                                                 enriched=args.enriched, runs=args.runs, distance=args.distance_measure,
                                                 background_model=args.background_model, replace=args.replace,
                                                 network_data={"network_file": args.network,
                                                               "prop_name": args.network_property_name,
                                                               "id_type": args.network_id_type})
            ru.print_current_usage('Save files') if args.verbose else None
            save_contribution_results(results=res_sig, prefix=pref, out_dir=args.out_dir)
            if args.plot:
                ru.print_current_usage('Save plots') if args.verbose else None
                i_type = "genes" if args.target_id_type in ['uniprot', 'entez', 'symbol', 'ensemble'] else "diseases"
                pu.create_contribution_plots(result_sig=res_sig, input_type=i_type, out_dir=args.out_dir, prefix=pref)

                if args.mode in ['subnetwork', 'subnetwork-set']:
                    pu.create_contribution_graphs(result_sig=res_sig, input_type=i_type,
                                                  network_data={"network_file": args.network,
                                                                "prop_name": args.network_property_name,
                                                                "id_type": args.network_id_type},
                                                  out_dir=args.out_dir, prefix=pref)



import pandas as pd
from .mappers import mapping_utils as mu
from .mappers.mapper import Mapper
from . import config as c
from abc import abstractmethod
import random
# only in full biodigest
import graph_tool as gt
import graph_tool.util as gtu
import graph_tool.topology as gtt
from graph_tool import GraphView


class BackgroundModel():

    @abstractmethod
    def get_module(self, **kwargs):
        pass


class CompleteModel(BackgroundModel):

    def __init__(self, prev_id_type, full_id_map):
        self.full_id_set = full_id_map[full_id_map[c.ID_TYPE_KEY[prev_id_type]] != ""][
            c.ID_TYPE_KEY[prev_id_type]].tolist()

    def get_module(self, to_replace):
        random_sample = set(random.sample(self.full_id_set, len(to_replace)))
        return random_sample


class TermPresModel(BackgroundModel):

    def __init__(self, mapper: Mapper, prev_id_type, new_id_type, map_id_type, map_att_type, term):
        # prepare candidates
        att_map = mu.map_to_prev_id(main_id_type=c.ID_TYPE_KEY[new_id_type],
                                    id_type=c.ID_TYPE_KEY[prev_id_type],
                                    id_mapping=mapper.loaded_mappings[map_id_type],
                                    att_mapping=mapper.loaded_mappings[map_att_type])
        self.atts_to_size(pd_map=att_map)
        self.size_mapping_to_dict(pd_size_map=self.att_len, id_col=c.ID_TYPE_KEY[prev_id_type], term_col=term,
                                  threshold=100)

    def get_module(self, to_replace, term, prev_id_type):
        random_sample = set()
        for replace_id in to_replace:
            if replace_id in self.size_mapping:  # only if id is mappable to other ids
                random_sample.add(
                    self.att_len[self.att_len[term].isin(self.size_mapping[replace_id])][
                        c.ID_TYPE_KEY[prev_id_type]].sample(
                        n=1).values[0])
        return random_sample

    def atts_to_size(self, pd_map: pd.DataFrame):
        att_len = pd_map.copy()
        att_len[att_len.columns[1:]] = att_len[att_len.columns[1:]].applymap(mu.set_to_len)
        att_len['sum'] = att_len[att_len.columns[1:]].sum(axis=1)
        self.att_len = att_len

    def size_mapping_to_dict(self, pd_size_map: pd.DataFrame, id_col: str, term_col: str, threshold: int = 100):
        size_to_occ = pd.DataFrame(pd_size_map[term_col].value_counts()).sort_index().to_dict()[term_col]
        pd_size_map = pd_size_map.sort_values(by=[id_col]).reset_index(drop=True)
        new_dict = dict()
        term_sizes = pd_size_map[term_col].unique().tolist()
        for index, key in enumerate(term_sizes):
            curr_keys = [key]
            if size_to_occ[key] < threshold:
                sum_tmp, add_top, add_bottom = size_to_occ[key], index, index
                while sum_tmp < threshold:
                    if add_top - 1 >= 0:
                        add_top = add_top - 1
                        sum_tmp = sum_tmp + size_to_occ[term_sizes[add_top]]
                        curr_keys.append(term_sizes[add_top])
                    if add_bottom + 1 < len(term_sizes):
                        add_bottom = add_bottom + 1
                        sum_tmp = sum_tmp + size_to_occ[term_sizes[add_bottom]]
                        curr_keys.append(term_sizes[add_bottom])
            for cur_id in pd_size_map[pd_size_map[term_col] == key][id_col]:
                new_dict[cur_id] = curr_keys
        self.size_mapping = new_dict


class NetworkModel(BackgroundModel):

    def __init__(self, network_data: dict, to_replace, N):
        self.network_type = network_data['id_type']
        if network_data["network_file"].endswith(".sif"):
            df = pd.read_csv(network_data["network_file"], sep="\t", header=None)
            G = gt.Graph(directed=False)
            v_ids = G.add_edge_list(df[[0, 2]].values, hashed=True)
            G.vertex_properties['id'] = v_ids
        else:
            G = gt.load_graph(network_data["network_file"])
            G.vertex_properties['id'] = G.vertex_properties[network_data['prop_name']]
        # Reduce to ids that are present in the network
        to_replace_filtered = to_replace.intersection(set(G.vertex_properties['id']))
        # Find number of CCs for the input module
        ccs_num = self.find_num_ccs(G, to_replace_filtered)
        # Generate random modules of matched number of connected components
        self.generate_rand_modules(G, N, ccs_num, len(to_replace_filtered))
        self.node_ids = {node: G.vertex_properties["id"][node] for node in range(G.num_vertices())}

    def get_module(self, index):
        return set([self.node_ids[node] for node in self.rand_module_nodes_list[index]])

    def find_num_ccs(self, G, module_nodes):
        G.set_directed(False)
        module_nodes_ids = [int(gtu.find_vertex(G, prop=G.vertex_properties['id'], match=node)[0]) for node
                            in module_nodes]
        v_modul_filt = G.new_vertex_property('bool')
        for i in module_nodes_ids:
            v_modul_filt[i] = True

        induced = GraphView(G, v_modul_filt)
        ees = [(G.vertex_index[e.source()], G.vertex_index[e.target()]) for e in induced.edges()]
        induced_module = gt.Graph(directed=False)
        node_id_to_node = {node_id: induced_module.add_vertex() for node_id in module_nodes}
        node_id_property = induced_module.new_vp('string', vals=module_nodes)
        induced_module.vertex_properties['id'] = node_id_property
        edges = []
        for e in ees:
            source = node_id_to_node[G.vertex_properties['id'][e[0]]]
            target = node_id_to_node[G.vertex_properties['id'][e[1]]]
            edges.append((source, target))
        induced_module.add_edge_list(edges)

        comp, hist = gtt.label_components(induced_module, directed=None, attractors=False)
        comp_labels = [comp.a[node] for node in range(induced_module.num_vertices())]
        ccs_num = max(comp_labels) + 1
        return ccs_num

    def generate_rand_modules(self, G, N, ccs_num, MS):
        random_cc_modules_list = []
        rand_module_nodes_list = []
        for r in range(N):
            random_cc_module, rand_module_nodes, iterr = self._rand_module(G, ccs_num, MS)
            if iterr <= 20:
                random_cc_modules_list.append(random_cc_module)
                rand_module_nodes_list.append(rand_module_nodes)
            else:
                random_cc_module, rand_module_nodes, iterr = self._rand_module(G, ccs_num, MS)
                random_cc_modules_list.append(random_cc_module)
                rand_module_nodes_list.append(rand_module_nodes)
            if len(rand_module_nodes) == 0:
                print("tja...")
        self.rand_module_nodes_list = rand_module_nodes_list


    def _rand_module(self, G, ccs_num, MS):
        # step 1: randomly select ccs_num number of seeds (which are not neighbors)
        G_nodes = list(range(G.num_vertices()))
        random_cc_module = dict()
        initial_seeds = []
        neighb_initial_seeds = []
        for c in range(ccs_num):
            added = False
            while not added:
                s = random.choice(G_nodes)

                if s not in initial_seeds and s not in neighb_initial_seeds and len(G.get_all_neighbors(s)>0):
                    added = True
                    initial_seeds.append(s)
                    neighb_initial_seeds.extend(list(G.get_all_neighbors(s)))
                    random_cc_module[s] = []

        rand_module_nodes = set()
        rand_module_nodes.update(initial_seeds)
        # step 2: expand the selected seeds by their neighbors enforcing the number of CCs remain the same
        for cs in initial_seeds:
            #print(G.get_all_neighbors(cs))
            #print(cs)
            rn = random.choice(G.get_all_neighbors(cs))
            intersect = False
            for k in random_cc_module.keys():
                if k != cs and (len(set(G.get_all_neighbors(rn)).intersection(
                        set(random_cc_module[k]))) > 0 or rn in G.get_all_neighbors(k)):
                    intersect = True
                    break
            if not intersect:
                random_cc_module[cs].append(rn)
                rand_module_nodes.add(rn)
            if len(rand_module_nodes) >= MS:
                break
        iterr = 0
        while len(rand_module_nodes) < MS and iterr < 20:
            iterr += 1
            for k in initial_seeds:
                vv = random_cc_module[k]
                # for k, vv in random_cc_modules.items():
                for v in vv:
                    rn = random.choice(G.get_all_neighbors(v))
                    if rn not in vv:
                        intersect = False
                        for i in random_cc_module.keys():
                            if i != k and (len(set(G.get_all_neighbors(rn)).intersection(
                                    set(random_cc_module[i]))) > 0 or rn in G.get_all_neighbors(i)):
                                intersect = True
                                break
                        if not intersect:
                            random_cc_module[k].append(rn)
                            rand_module_nodes.add(rn)
                    if len(rand_module_nodes) >= MS:
                        break
                if len(rand_module_nodes) >= MS:
                    break
        return random_cc_module, rand_module_nodes, iterr